@@ -4,7 +4,7 @@ Copyright 2021 Objectiv B.V.
 import itertools
 from copy import copy
 from enum import Enum
-from typing import Union, List, Tuple, Optional, Dict, Set, Hashable, cast, NamedTuple, Sequence
+from typing import Union, List, Tuple, Optional, Dict, Set, Hashable, cast, NamedTuple, Sequence, Mapping
 
 from sqlalchemy.engine import Dialect
 
@@ -14,7 +14,8 @@ from bach.dataframe import DtypeNamePair
 from bach.expression import Expression, join_expressions, TableColumnReferenceToken, \
     NonAtomicExpression, ExpressionToken, ColumnReferenceToken
 from bach.utils import ResultSeries, get_result_series_dtype_mapping
-from sql_models.model import Materialization, CustomSqlModelBuilder, SqlModel
+from sql_models.constants import NotSet, not_set
+from sql_models.model import Materialization, CustomSqlModelBuilder, SqlModel, SqlModelSpec
 from bach.sql_model import BachSqlModel, construct_references
 
 
@@ -427,18 +428,28 @@ def _determine_series_per_source(
     right_node = cast(BachSqlModel, base.base_node.references['right_node'])
 
     expressions_to_parse: Dict[str, List[Expression]] = {}
+
+    merge_node = cast(MergeSqlModel, base.base_node)
+    merge_on = merge_node.merge_on
+
+    coalesced_series = [
+        left_on for left_on, right_on in zip(merge_on.left, merge_on.right) if left_on == right_on
+    ]
     for series_name, original_expr in base.base_node.column_expressions.items():
         if series_name not in series_to_unmerge:
             continue
 
-        if isinstance(original_expr, NonAtomicExpression) or not original_expr.has_table_column_references:
+        if series_name not in coalesced_series:
             expressions_to_parse[series_name] = [original_expr]
             continue
 
+        # removes first-level tokens from COALESCE expression
+        # I.e COALESCE(expr1, expr2)
+        # expression_to_parse[series] = [expr1, expr2]
         expressions_to_parse[series_name] = [
-            sub_expr if isinstance(sub_expr, Expression) else Expression([sub_expr])
+            sub_expr
             for sub_expr in original_expr.data
-            if isinstance(sub_expr, (Expression, TableColumnReferenceToken))
+            if isinstance(sub_expr, Expression)
         ]
 
     default_series = base.all_series[base.data_columns[0]]
@@ -514,9 +525,10 @@ def _get_merge_sql_model(
         columns_expr=columns_expr,
         join_type_expr=join_type_expr,
         on_clause=on_clause,
+        merge_on=merge_on,
         left_node=left.base_node,
         right_node=right.base_node,
-        variables=variables
+        variables=variables,
     )
 
 
@@ -583,7 +595,14 @@ def _resolve_merge_expression_references(
                     df_series = right
                     table_alias = 'r'
 
-                resolved_expr = _get_expression(df_series=df_series, label=nested_token.column_name)
+                # nested_token.column_name might not be the same to token.column_name
+                # since series can be renamed
+                expr_label = (
+                    nested_token.column_name
+                    if nested_token.column_name == cond.name
+                    else token.column_name
+                )
+                resolved_expr = _get_expression(df_series=df_series, label=expr_label)
                 resolved_expr = resolved_expr.resolve_column_references(dialect, table_alias)
                 resolved_nested_tokens.append(resolved_expr)
 
@@ -606,9 +625,14 @@ def _get_expression(df_series: DataFrameOrSeries, label: str) -> Expression:
 
 
 class MergeSqlModel(BachSqlModel):
-    def __init__(self, *args, **kwargs):
+    def __init__(self, merge_on: MergeOn, *args, **kwargs):
+        self._merge_on = merge_on
         super().__init__(*args,  **kwargs)
 
+    @property
+    def merge_on(self) -> MergeOn:
+        return self._merge_on
+
     @classmethod
     def get_instance(
         cls,
@@ -618,6 +642,7 @@ class MergeSqlModel(BachSqlModel):
         columns_expr: Expression,
         join_type_expr: Expression,
         on_clause: Expression,
+        merge_on: MergeOn,
         left_node: BachSqlModel,
         right_node: BachSqlModel,
         variables: Dict['DtypeNamePair', Hashable],
@@ -658,4 +683,32 @@ class MergeSqlModel(BachSqlModel):
             materialization=Materialization.CTE,
             materialization_name=None,
             column_expressions=column_expressions,
+            merge_on=merge_on,
+        )
+
+    def copy_override(
+        self: 'MergeSqlModel',
+        *,
+        model_spec: SqlModelSpec = None,
+        placeholders: Mapping[str, Hashable] = None,
+        references: Mapping[str, 'SqlModel'] = None,
+        materialization: Materialization = None,
+        materialization_name: Union[Optional[str], NotSet] = not_set,
+        column_expressions: Dict[str, Expression] = None,
+        merge_on: MergeOn = None
+    ) -> 'MergeSqlModel':
+        """
+        Similar to super class's implementation, but adds optional 'merge_on' parameter
+        """
+        materialization_name_value = (
+            self.materialization_name if materialization_name is not_set else materialization_name
+        )
+        return self.__class__(
+            model_spec=self.model_spec if model_spec is None else model_spec,
+            placeholders=self.placeholders if placeholders is None else placeholders,
+            references=self.references if references is None else references,
+            materialization=self.materialization if materialization is None else materialization,
+            materialization_name=materialization_name_value,
+            column_expressions=self.column_expressions if column_expressions is None else column_expressions,
+            merge_on=self.merge_on if merge_on is None else merge_on
         )
