@@ -1,7 +1,7 @@
 """
 Copyright 2021 Objectiv B.V.
 """
-from bach.series.series import Series, const_to_series
+from bach.series.series import Series, value_to_series
 from bach.series.series_numeric import \
     SeriesAbstractNumeric, SeriesInt64, SeriesFloat64
 from bach.series.series_boolean import SeriesBoolean
@@ -10,3 +10,6 @@ from bach.series.series_json import SeriesJson, SeriesJsonb
 from bach.series.series_string import SeriesString
 from bach.series.series_datetime import \
     SeriesAbstractDateTime, SeriesDate, SeriesTime, SeriesTimestamp, SeriesTimedelta
+from bach.series.series_multi_level import SeriesAbstractMultiLevel, SeriesNumericInterval
+from bach.series.series_list import SeriesList
+from bach.series.series_dict import SeriesDict
