@@ -1,16 +1,18 @@
 """
 Copyright 2021 Objectiv B.V.
 """
-from typing import Union, TYPE_CHECKING
+from typing import Union, TYPE_CHECKING, Optional
 
 from sqlalchemy.engine import Dialect
 
 from bach.series import Series
 from bach.expression import Expression
+from bach.types import StructuredDtype
 from sql_models.constants import DBDialect
 
 if TYPE_CHECKING:
     from bach.series import SeriesBoolean
+    from bach import DataFrame
 
 
 class StringOperation:
@@ -23,7 +25,7 @@ class StringOperation:
         Get a python string slice using DB functions. Format follows standard slice format
         Note: this is called 'slice' to not destroy index selection logic
         :param item: an int for a single character, or a slice for some nice slicing
-        :return: BuhTuhSeriesString with the slice applied
+        :return: SeriesString with the slice applied
         """
         if isinstance(start, (int, type(None))):
             item = slice(start, start + 1)
@@ -124,7 +126,12 @@ class SeriesString(Series):
         return literal
 
     @classmethod
-    def supported_value_to_literal(cls, dialect: Dialect, value: str) -> Expression:
+    def supported_value_to_literal(
+        cls,
+        dialect: Dialect,
+        value: str,
+        dtype: StructuredDtype
+    ) -> Expression:
         return Expression.string_value(value)
 
     @classmethod
@@ -133,6 +140,31 @@ class SeriesString(Series):
             return expression
         return Expression.construct(f'cast({{}} as {cls.get_db_dtype(dialect)})', expression)
 
+    def get_dummies(
+        self,
+        prefix: Optional[str] = None,
+        prefix_sep: str = '_',
+        dummy_na: bool = False,
+        dtype: str = 'int64',
+    ) -> 'DataFrame':
+        """
+        Convert each unique category/value from the series into a dummy/indicator variable.
+
+        :param prefix: String to append to each new column name. By default, the prefix will be the name of
+            the caller.
+        :param prefix_sep: Separated between the prefix and label.
+        :param dummy_na: If true, it will include ``nan`` as a variable.
+        :param dtype: dtype of all new columns
+
+        :return: DataFrame
+
+        .. note::
+            Series should contain at least one index level.
+        """
+        return self.to_frame().get_dummies(
+            prefix=prefix, prefix_sep=prefix_sep, dummy_na=dummy_na, dtype=dtype,
+        )
+
     @property
     def str(self) -> StringOperation:
         """
